from pytest import raises
import operator as op
from typing import TypeAlias, Annotated
from contractme import annotated, precondition, postcondition
from annotated_types import Ge, Le
import hypothesis

from contractme.contracting import ContractedFunction
import contractme.testing
from contractme.testing import autotest, get_generator

"""
This contains tests for the testing extension.

As this is testing the testing code, it is probably harder to understand that
it needs to be.

Refer to the doc for simpler examples.
"""


MyInt: TypeAlias = Annotated[int, Ge(0), Le(1000)]


@annotated
def sub_and_div_no_contract(a: MyInt, b: MyInt) -> float:
    """
    This is incorrectly contracted: it fails for a == b
    """
    return 1.0 / float(a - b)


def test_trivial_failure_missing_contract():
    with raises(ZeroDivisionError):
        autotest(sub_and_div_no_contract)


@annotated
@precondition(lambda a, b: a != b)
def sub_and_div(a: MyInt, b: MyInt) -> float:
    """
    This is correctly contracted
    """
    return 1.0 / float(a - b)


def add(a: int, b: int) -> int:
    return a + b


def test_test_with_examples_non_contracted():
    gen = contractme.testing.test_with_examples(
        add,
        (1, 0),
        (3, 2),
        {"a": 4, "b": 10},
    )
    gen()


def test_no_failure():
    autotest(sub_and_div)


def test_div_force_fail():
    gen_sub_and_div = get_generator(sub_and_div)
    # kinda weird to have this double call, but that's decorators for you...
    gen = hypothesis.example((0, 0))(gen_sub_and_div)
    with raises(AssertionError):
        # catched by contracts at call
        gen()


def test_test_with_examples():
    gen = contractme.testing.test_with_examples(
        sub_and_div,
        (1, 0),
        (3, 2),
        {"a": 4, "b": 10},
    )
    gen()


def test_test_with_examples_contract_fails():
    with raises(RuntimeError):
        # catched by contracts at test generation
        _ = contractme.testing.test_with_examples(
            sub_and_div,
            (1, 0),
            (0, 0),
        )


def test_test_with_examples_function_fails():
    # no contracts, no catch
    gen = contractme.testing.test_with_examples(
        sub_and_div_no_contract,
        (1, 0),
        (0, 0),
    )

    with raises(ZeroDivisionError):
        gen()


@annotated
@precondition(lambda a, b: a != b)
@postcondition(lambda result: result > 0.0)
def sub_and_div_wrong_impl(a: MyInt, b: MyInt) -> float:
    """
    This is correctly contracted
    """
    return 1.0 / float(a - b)


def test_test_with_examples_postcondition_checked_at_call():
    # no contracts, no catch
    gen = contractme.testing.test_with_examples(
        sub_and_div_wrong_impl,
        (1, 10),
    )

    with raises(AssertionError):
        gen()



import sys
from typing import get_type_hints
from inspect import get_annotations
from contractme.contracting import precondition, postcondition
from contractme.typecheck import (
    get_first_structure_error,
    get_constraints_errors,
    resolve_type,
    AnnotationForm,
)


def get_types_and_annotations_errors(value, an):
    """
    Checks the structure and annotations of data provided compared to the annotations.

    The structure is checked first, then iif it is valid, the annotations are. This allows
    for a better error messaging: you first provide a correct structure, then correct values.
    """
    # we visit all of the possible nodes depth-first
    # for each one, we check its type, and we delegate any annotation check
    # then we loop over all these annotations, and we check them as well
    # this has a potentially big runtime cost for nested option or recursive
    # types.
    struct_err = get_first_structure_error(value, an)
    if struct_err:
        return f"structure is malformed: {struct_err}"
    else:
        constraints_err = get_constraints_errors(value, an)
        if constraints_err:
            return "structure ok, but value does not match constraints: " + ", ".join(
                constraints_err
            )
        else:
            return None


def annotated(f):
    """Decorate a function so that its ``Annotated`` type hints are enforced at runtime.

    For each parameter and the return value, ``@annotated`` auto-generates a
    ``@precondition`` (for parameters) or ``@postcondition`` (for the return value) that
    validates both the type structure and any ``annotated-types`` constraints
    (``Ge``, ``Gt``, ``Le``, ``Lt``, ``MultipleOf``, ``MinLen``, ``MaxLen``,
    ``Predicate``).  Supported type structures: plain types, ``Annotated``,
    ``TypeAlias``, ``list``, ``set``, ``tuple``, ``dict``, ``Union`` / ``|``,
    and recursive types.

    Validation is two-phase: structural correctness first (e.g. "is this a list?"),
    then constraint checking (e.g. "are all elements ≥ 0?").  Errors are printed to
    ``stderr`` before raising.  Unknown annotations are silently accepted
    (open-world assumption).  Type checking uses ``isinstance``, so ``bool`` passes an
    ``int`` check.

    Some types are not yet supported; see https://gitlab.com/leogermond/contractme/-/work_items/9

    Example::

        from typing import Annotated
        from annotated_types import Ge, Lt

        @annotated
        def clamp(v: Annotated[float, Ge(0.0), Lt(1.0)]) -> Annotated[float, Ge(0.0), Lt(1.0)]:
            return min(max(v, 0.0), 1.0 - 1e-9)
    """

    def check_annotated_arg(arg_name, arg_an: AnnotationForm):
        constraints = getattr(arg_an, "__metadata__", tuple())
        t = resolve_type(arg_an)
        if constraints:
            msg = f"{arg_name} should be instance of {t} under constraints {constraints}"
        else:
            msg = f"{arg_name} should be instance of {t}"

        if arg_name == "return":
            cond_arg_name = "result"
            cond_f = postcondition
        else:
            cond_arg_name = arg_name
            cond_f = precondition

        def chk(**kw):
            err = get_types_and_annotations_errors(kw[cond_arg_name], arg_an)
            if err:
                print(f"arg {cond_arg_name}: {err}", file=sys.stderr)

            return err is None

        cond = cond_f(chk, msg)
        return cond

    ths = get_type_hints(f)
    ans = get_annotations(f)

    for n, t_or_typevar in ths.items():
        f = check_annotated_arg(n, ans[n])(f)

    return f



import inspect
import copy
from warnings import warn
from typing import Callable
import functools
from typing import Callable, Any, Generic, TypeVar


def par_inner(s: str) -> str:
    """
    returns the "parenthesis inner member"
    which is the expression inside parentheses, if there are parentheses.
    """
    # pre
    assert s != "", "input cannot be empty"
    assert any(
        c not in "( )" for c in s
    ), f"input {s!r} must contain other characters than spaces and parentheses"

    # s = prelude + body
    #
    # where prelude contains spaces and opening par
    # and body contains all sorts of chars and closing par

    ipar = 0
    # inner parentheses, inside the body
    prelude = True
    # are we still in prelude vs body?
    # nb the "limit" char is processed as both prelude and body
    body_len = 0
    # number of chars processed in body
    r = s
    while body_len != len(r):
        if prelude:
            c = r[0]
            if c == "(":
                r = r[1:]
            elif c == " ":
                r = r[1:]
            else:
                # this belongs to ~a museum~ the body!
                prelude = False

        if not prelude:
            c = r[body_len]
            body_len += 1

            if c == "(":
                ipar += 1
            elif c == ")":
                if ipar > 0:
                    ipar -= 1
                else:
                    body_len -= 1
                    r = r[:body_len]

    r = r.rstrip(" ")

    # post
    assert r != ""
    assert r[0] not in "( "
    assert r[-1] not in " "
    assert r.count("(") == r.count(")")
    return r


def show_source(f: Callable) -> str:
    def get_fn_name(f: Callable) -> str:
        return f.__code__.co_name

    def get_source_for_lambda(s: str) -> str:
        slambda = s[s.index("lambda") + len("lambda") :]
        slamexpr = slambda[slambda.index(":") + 1 :]

        return par_inner(slamexpr)

    s = inspect.getsource(f)
    if s.lstrip().startswith("def "):
        return get_fn_name(f)
    elif "lambda" in s:
        return get_source_for_lambda(s)
    else:  # pragma: no cover
        # Not reachable?
        raise Exception(f"impossible to parse source {s}")


def get_wraparound(f):
    if hasattr(f, "contractme_wraparound"):
        result = (
            f.contractme_wraparound,
            f.contractme_preconditions,
            f.contractme_postconditions,
        )
    else:
        result = f, [], []
    return result


def assert_argspec_and_kw_match(argspec, kw, ignore=None):
    assert all(
        a in kw for a in argspec.args if not ignore or a not in ignore
    ), f"missing arg in {list(kw)} for {argspec.args}"


def get_kw_from_call(f, a, kw):
    argspec = inspect.getfullargspec(f)

    kw2 = kw.copy()
    # kwargs
    for i, nm in enumerate(argspec.args):
        # positionals
        if len(a) > i:
            v = a[i]
            kw2[nm] = v

    for i, nm in enumerate(a for a in argspec.args if a not in kw2):
        # defaults
        if argspec.defaults is not None and len(argspec.defaults) > i:
            kw2[nm] = argspec.defaults[i]
        else:
            # that's an error, but from the caller, let python handle that
            warn(
                f"call to {f} will fail due to wrong arguments:"
                f"missing {nm} in {kw2} ({argspec}), contracts ignored"
            )
            return {}

    assert_argspec_and_kw_match(argspec, kw2)
    return kw2


def format_kw(kw):
    def brief_repr(o, length):
        r = repr(o)
        if len(r) > length:
            half = 1 + ((length - 5 + 1) // 2)
            r = f"{r[:half]}[...]{r[-half:]}"
        return r

    return ", ".join(f"{k} = {brief_repr(kw[k], 30)}" for k in sorted(kw))


class OldHolder:
    def __init__(self, d):
        self.d = copy.deepcopy(d)

    def __getattr__(self, name):
        return self.d[name]

    def __repr__(self):
        return repr(self.d)


PRE, PRE_INIT, POST = 1, 2, 3

checked = {PRE, POST}


def is_checked(cond_type):
    assert cond_type in (PRE, POST)
    return __debug__ and cond_type in checked


def check(cond_type):
    assert cond_type in (PRE, POST)
    assert not is_checked(cond_type), f"{cond_type} is already checked"
    checked.add(cond_type)


def ignore(cond_type):
    assert cond_type in (PRE, POST)
    assert is_checked(cond_type), f"{cond_type} is already ignored"
    checked.remove(cond_type)


def ignore_preconditions():
    ignore(PRE)


def check_preconditions():
    check(PRE)


def ignore_postconditions():
    ignore(POST)


def check_postconditions():
    check(POST)


def get_kw_for_conditional_before(cond_type, conditional, in_kw):
    assert cond_type in (PRE, PRE_INIT, POST)
    assert callable(conditional)
    out_kw = {}
    argspec = inspect.getfullargspec(conditional)

    if argspec.varargs:  # pragma: no cover
        raise NotImplementedError("*args in conditions")

    if argspec.varkw is not None:
        if argspec.args:  # pragma: no cover
            raise NotImplementedError("mixing args and **kw in conditions")

        out_kw = in_kw
    else:
        for nm in argspec.args:
            if nm == "result":
                assert cond_type == POST, f"only postconditions can use {nm!r}"
            elif nm == "old":
                assert cond_type == POST, f"only postconditions can use {nm!r}"
                out_kw[nm] = OldHolder(in_kw)
            else:
                out_kw[nm] = in_kw[nm]

    assert_argspec_and_kw_match(argspec, out_kw, ("result",))
    return out_kw


def get_kw_for_postcondition_after(result, conditional, in_kw):
    assert callable(conditional)
    out_kw = {}
    argspec = inspect.getfullargspec(conditional)
    must_have_result = "result" in argspec.args or argspec.varkw is not None
    if must_have_result:
        out_kw["result"] = result
    assert ("result" in out_kw) == must_have_result
    return out_kw


F_Ret = TypeVar("F_Ret")
type F[F_Ret] = Callable[..., F_Ret]
type F_Pred = Callable[..., bool]


class ContractedFunction[F_Ret]:
    def __init__(
        self,
        call_and_run_contracts: F,
        original_call: F,
        preconditions: list[F],
        postconditions: list[F],
        accepts: F_Pred,
        rejects: F_Pred,
    ):
        functools.update_wrapper(self, original_call)
        self.call_and_run_contracts = call_and_run_contracts
        self.contractme_wraparound = original_call
        self.contractme_preconditions = preconditions
        self.contractme_postconditions = postconditions
        self.accepts = accepts
        self.rejects = rejects
        self.is_method = False

    def __get__(self, instance, owner):
        if not self.is_method:
            self.is_method = True
        self.instance = instance

        return self

    def __call__(self, *a, **kw) -> F_Ret:
        return self.call_and_run_contracts(self, *a, **kw)


type Any_Callable[F_Ret] = Callable[..., F_Ret]
type Predicate_F = Callable[..., bool]


def condition_decorator[F_Ret](
    is_precondition: bool, conditional: Predicate_F, message: str | None = None
) -> Callable[[Any_Callable[F_Ret]], ContractedFunction[F_Ret]]:
    def deco(f: Callable):
        if is_precondition and f.__name__ == "__init__":
            raise RuntimeError("__init__ cannot have precondition")
        original_call, preconditions, postconditions = get_wraparound(f)

        if is_precondition:
            preconditions.insert(0, (conditional, message))
        else:
            postconditions.insert(0, (conditional, message))

        def rejects_call_kw(call_kw, stop_at_first_rejection: bool = False) -> list:
            rejection_reasons = []
            for precondition, message in preconditions:
                cond_type = PRE_INIT if original_call.__name__ == "__init__" else PRE
                cond_kw = get_kw_for_conditional_before(cond_type, precondition, call_kw)
                if not precondition(**cond_kw):
                    cond_kw_str = format_kw(cond_kw)
                    src = show_source(precondition)
                    if message is None:
                        rejection_reasons.append(
                            f"**caller** of {original_call.__name__}() bug:"
                            f"precondition {src!r} failed: {cond_kw_str}"
                        )
                    else:
                        rejection_reasons.append(
                            f"**caller** of {original_call.__name__}() bug:"
                            f"precondition {message!r} failed: {cond_kw_str}"
                        )

                    if stop_at_first_rejection:
                        break

            return rejection_reasons

        def rejects(*a, **kw) -> bool:
            call_kw = get_kw_from_call(original_call, a, kw)
            return bool(rejects_call_kw(call_kw, stop_at_first_rejection=True))

        def accepts(*a, **kw) -> bool:
            call_kw = get_kw_from_call(original_call, a, kw)
            return not rejects_call_kw(call_kw, stop_at_first_rejection=True)

        def precondition_check(call_kw):
            reasons = rejects_call_kw(call_kw, stop_at_first_rejection=False)
            if reasons:
                raise AssertionError(reasons[0])

        def postcondition_prepare(call_kw):
            post_kw = []
            for postcondition, _ in postconditions:
                cond_kw = get_kw_for_conditional_before(POST, postcondition, call_kw)
                post_kw.append(cond_kw)
            return post_kw

        def postcondition_check(result, post_kw, call_kw):
            for i, (postcondition, message) in enumerate(postconditions):
                cond_kw = post_kw[i]
                cond_kw.update(get_kw_for_postcondition_after(result, postcondition, call_kw))
                if not postcondition(**cond_kw):
                    cond_kw_str = format_kw(cond_kw)
                    src = show_source(postcondition)
                    if message is None:
                        raise AssertionError(
                            f"{original_call.__name__}() bug: "
                            f"postcondition {src!r} failed: {cond_kw_str}"
                        )
                    else:
                        raise AssertionError(
                            f"{original_call.__name__}() bug: "
                            f"postcondition {message!r} failed: {cond_kw_str}"
                        )

        def call_and_run_contracts(f: ContractedFunction, *a, **kw):
            if f.is_method:
                a = [f.instance] + [*a]

            call_kw = get_kw_from_call(original_call, a, kw)

            if is_checked(PRE):
                precondition_check(call_kw)

            post_kw = postcondition_prepare(call_kw) if is_checked(POST) else None

            result = original_call(*a, **kw)

            if is_checked(POST):
                postcondition_check(result, post_kw, call_kw)

            return result

        return ContractedFunction(
            call_and_run_contracts,
            original_call,
            preconditions,
            postconditions,
            accepts,
            rejects,
        )

    return deco


def precondition(conditional: Predicate_F, message: str | None = None):
    return condition_decorator(True, conditional, message)


def postcondition(conditional: Predicate_F, message: str | None = None):
    return condition_decorator(False, conditional, message)


never_returns = postcondition(lambda: False, "should never return")



"""Type stubs for contractme library."""

from typing import Callable, TypeVar, Any, overload
from typing_extensions import ParamSpec

F = TypeVar("F", bound=Callable[..., Any])
P = ParamSpec("P")
R = TypeVar("R")

@overload
def precondition(
    condition: Callable[[Any], bool],
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def precondition(
    condition: Callable[[Any], bool], str
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def precondition(condition: Callable[P, bool]) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def precondition(
    condition: Callable[P, bool], str
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def precondition(
    condition: Callable[..., bool],
) -> Callable[[Callable[..., R]], Callable[..., R]]: ...
@overload
def precondition(
    condition: Callable[..., bool], str
) -> Callable[[Callable[..., R]], Callable[..., R]]: ...
@overload
def postcondition(
    condition: Callable[[Any], bool],
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def postcondition(
    condition: Callable[[Any], bool], str
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def postcondition(
    condition: Callable[[Any, Any], bool],
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def postcondition(
    condition: Callable[[Any, Any], bool], str
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...
@overload
def postcondition(
    condition: Callable[..., bool],
) -> Callable[[Callable[..., R]], Callable[..., R]]: ...
@overload
def postcondition(
    condition: Callable[..., bool], str
) -> Callable[[Callable[..., R]], Callable[..., R]]: ...



from contractme.contracting import ContractedFunction
from typing import Callable, get_type_hints, TypeAlias
from hypothesis import assume, given, example, strategies as st


def get_strategies(f: Callable) -> list[st.SearchStrategy]:
    th = get_type_hints(f, include_extras=True)

    sts = []

    for n, t in th.items():
        if n == "return":
            continue
        sts.append(st.from_type(t))
    return sts


TestableFunction: TypeAlias = ContractedFunction | Callable
TestGenerator: TypeAlias = Callable


def call_with_composite(f: Callable, composite: tuple | dict):
    if isinstance(composite, dict):
        return f(**composite)
    else:
        return f(*composite)


def draw_accepted_input(draw: Callable, f: TestableFunction) -> tuple:
    sts = get_strategies(f)
    v = tuple(draw(e) for e in sts)
    if isinstance(f, ContractedFunction):
        assume(f.accepts(*v))
    return v


def get_generator(f: TestableFunction) -> TestGenerator:
    @st.composite
    def correct_input(draw: Callable) -> tuple[int, int]:
        return draw_accepted_input(draw, f)

    @given(correct_input())
    def search(value):
        call_with_composite(f, value)

    return search


def autotest(f: TestableFunction):
    get_generator(f)()


def test_with_examples(f: TestableFunction, *exs) -> TestGenerator:
    gen = get_generator(f)
    for e in exs:
        if isinstance(f, ContractedFunction):
            accepted = call_with_composite(f.accepts, e)
            if not accepted:
                raise RuntimeError(f"contracts rejected example {e}")
        gen = example(e)(gen)
    return gen



from typing import Annotated
from annotated_types import Ge, Lt, Gt
from inspect import get_annotations
import dataclasses
from contractme.typecheck import (
    get_recursive_composite_type_info,
    get_first_structure_error,
    get_constraints_errors,
    ChildMod,
)

# these are absurd bottom types with dummy annotations
# used for checking low-level recursivity-related function
type TR2 = TR  # type: ignore
type TR = TR2
type T0 = Annotated[TR, 3]
type T = Annotated[T0, 2]
type T2 = tuple[int, T]
type T3 = Annotated[T2, 1]


def f[TT: T3](a: TT) -> TT:  # pragma: no cover
    return a


def g(a: tuple[int, ...]):  # pragma: no cover
    pass


def h(a: int | Annotated[float, Ge(0.0)]):  # pragma: no cover
    pass


def test_get_info_complex():
    an = get_annotations(f)["a"]
    type_infos = [dataclasses.astuple(e) for e in get_recursive_composite_type_info(an)]
    assert type_infos == [
        (True, tuple, [1], ChildMod.NONE, (int, T)),
        (True, int, [], ChildMod.NONE, tuple()),
        (False, TR, [2, 3], ChildMod.NONE, tuple()),
    ]


def test_get_info_tuple_ellipsis():
    an = get_annotations(g)["a"]
    type_infos = [dataclasses.astuple(e) for e in get_recursive_composite_type_info(an)]
    assert type_infos == [(True, tuple, [], ChildMod.REPEAT, (int,))]


def test_structure_error_union():
    an = get_annotations(h)["a"]
    assert get_first_structure_error(1, an) is None

    # tricky: constraint error but structure is OK
    assert get_first_structure_error(-1.0, an) is None

    assert get_first_structure_error([], an) is not None
    assert get_first_structure_error("string", an) is not None


def test_constraints_error_union():
    an = get_annotations(h)["a"]
    assert not get_constraints_errors(0.0, an)
    assert get_constraints_errors(-1.0, an)


# ((<0.0) x (int)) + ((>0.0) x ((<0.0) x (int)) + (>0.0)))
type PosFloat = Annotated[float, Gt(0.0)]
type NegFloat = Annotated[float, Lt(0.0)]
type NegAndInt = tuple[NegFloat, int]
type NegAndIntOrPos = NegAndInt | PosFloat
type Complex = NegAndInt | tuple[PosFloat, NegAndIntOrPos]


def recursive_union(a: Complex):  # pragma: no cover
    pass


def test_complex():
    an = get_annotations(recursive_union)["a"]
    assert not get_constraints_errors((-1.0, 1), an)
    assert get_constraints_errors((1.0, 1), an)
    assert not get_constraints_errors((1.0, (-1.0, 1)), an)
    assert get_constraints_errors((1.0, (1.0, 1)), an)
    assert not get_constraints_errors((1.0, 1.0), an)



import pytest
from contractme import annotated
from typing import Annotated, TypeAlias, assert_never
from annotated_types import (
    Lt,
    Le,
    Gt,
    Ge,
    Interval,
    MultipleOf,
    MinLen,
    MaxLen,
    Len,
    Timezone,
    doc,  # type: ignore
    # for some reason this triggers pyright?
    # predicates
    LowerCase,
)
import functools
import operator
from datetime import datetime, timezone
from zoneinfo import ZoneInfo


def test_int():
    @annotated
    def birthday(age: int):
        return age + 1

    assert birthday(1) == 2
    with pytest.raises(AssertionError):
        birthday(1.0)  # type: ignore
        # mistyped on purpose


def test_return_int():
    @annotated
    def birthday(age) -> int:
        return age + 1

    assert birthday(1) == 2
    with pytest.raises(AssertionError):
        birthday(1.0)


def test_bound_typevar():
    @annotated
    def add[I: int](age: I, added: I) -> int:
        return int(age + added)

    assert add(1, 1) == 2
    with pytest.raises(AssertionError):
        add(1.0, 1.0)  # type: ignore
        # mistyped on purpose


Age = Annotated[int, Ge(1)]


def test_precondition():
    @annotated
    def birthday(age: Age) -> Age:
        return age + 1

    assert birthday(1) == 2

    with pytest.raises(AssertionError):
        birthday(0)

    with pytest.raises(AssertionError):
        birthday(1.0)  # type: ignore
        # mistyped on purpose


def test_postcondition():
    @annotated
    def antibirthday(age: Age) -> Age:
        return age - 1

    with pytest.raises(AssertionError):
        antibirthday(1)


def test_postcondition_int():
    @annotated
    def birthday(age) -> Age:
        return age + 1.0

    with pytest.raises(AssertionError):
        birthday(1)


Lt0: TypeAlias = Annotated[int, Lt(0)]


def test_Lt():
    @annotated
    def f(lt0: Lt0):
        pass

    f(-1)

    with pytest.raises(AssertionError):
        f(0)


Le0: TypeAlias = Annotated[int, Le(0)]


def test_Le():
    @annotated
    def f(le0: Le0):
        pass

    f(0)

    with pytest.raises(AssertionError):
        f(1)


Ge0: TypeAlias = Annotated[int, Ge(0)]


def test_Ge():
    @annotated
    def f(ge0: Ge0):
        pass

    f(0)

    with pytest.raises(AssertionError):
        f(-1)


Gt0: TypeAlias = Annotated[int, Gt(0)]


def test_Gt():
    @annotated
    def f(gt0: Gt0):
        pass

    f(1)

    with pytest.raises(AssertionError):
        f(0)


Lt0_partial: TypeAlias = Annotated[int, functools.partial(operator.lt, 0)]


def test_functools_partial_Lt():
    @annotated
    def f(_: Lt0_partial):
        pass

    f(-1)

    with pytest.raises(AssertionError):
        f(0)


Le0_partial: TypeAlias = Annotated[int, functools.partial(operator.le, 0)]


def test_functools_partial_Le():
    @annotated
    def f(_: Le0_partial):
        pass

    f(0)

    with pytest.raises(AssertionError):
        f(1)


Ge0_partial: TypeAlias = Annotated[int, functools.partial(operator.ge, 0)]


def test_functools_partial_Ge():
    @annotated
    def f(_: Ge0_partial):
        pass

    f(0)

    with pytest.raises(AssertionError):
        f(-1)


Gt0_partial: TypeAlias = Annotated[int, functools.partial(operator.gt, 0)]


def test_functools_partial_Gt():
    @annotated
    def f(gt0: Gt0_partial):
        pass

    f(1)

    with pytest.raises(AssertionError):
        f(0)


class MyInt_gt:
    def __init__(self, val):
        self.val = val

    def __gt__(self, other: "MyInt_gt"):
        return self.val > other.val


Gt0_MyInt: TypeAlias = Annotated[MyInt_gt, Gt(MyInt_gt(0))]


def test_Gt_custom():
    @annotated
    def f_Gt0(gt0: Gt0_MyInt):
        pass

    f_Gt0(MyInt_gt(1))

    with pytest.raises(AssertionError):
        f_Gt0(MyInt_gt(0))


class MyInt_lt:
    def __init__(self, val):
        self.val = val

    def __lt__(self, other: "MyInt_lt"):
        return self.val < other.val


Lt0_MyInt: TypeAlias = Annotated[MyInt_lt, Lt(MyInt_lt(0))]


def test_Lt_custom():
    @annotated
    def f_Lt0(lt0: Lt0_MyInt):
        pass

    f_Lt0(MyInt_lt(-1))

    with pytest.raises(AssertionError):
        f_Lt0(MyInt_lt(0))


class MyInt_ge:
    def __init__(self, val):
        self.val = val

    def __ge__(self, other: "MyInt_ge"):
        return self.val >= other.val


Ge0_MyInt: TypeAlias = Annotated[MyInt_ge, Ge(MyInt_ge(0))]


def test_Ge_custom():
    @annotated
    def f_Ge0(ge0: Ge0_MyInt):
        pass

    f_Ge0(MyInt_ge(0))

    with pytest.raises(AssertionError):
        f_Ge0(MyInt_ge(-1))


Vacuous: TypeAlias = Annotated[int, Interval()]


def test_vacuous_Interval():
    @annotated
    def f(_: Vacuous):
        pass

    f(0)


Gt0_Interval: TypeAlias = Annotated[int, Interval(gt=0)]


def test_Interval_only_gt():
    @annotated
    def f(_: Gt0_Interval):
        pass

    f(1)
    f(1000)

    with pytest.raises(AssertionError):
        f(0)


Gt0Lt2: TypeAlias = Annotated[int, Interval(gt=0, lt=2)]


def test_Interval_gt_lt():
    @annotated
    def f(_: Gt0Lt2):
        pass

    f(1)

    with pytest.raises(AssertionError):
        f(0)
    with pytest.raises(AssertionError):
        f(2)


Ge0Lt2: TypeAlias = Annotated[int, Interval(ge=0, lt=2)]


def test_Interval_ge_lt():
    @annotated
    def f(_: Ge0Lt2):
        pass

    f(0)
    f(1)

    with pytest.raises(AssertionError):
        f(-1)
    with pytest.raises(AssertionError):
        f(2)


Gt0Le2: TypeAlias = Annotated[int, Interval(gt=0, le=2)]


def test_Interval_gt_le():
    @annotated
    def f(_: Gt0Le2):
        pass

    f(1)
    f(2)

    with pytest.raises(AssertionError):
        f(0)
    with pytest.raises(AssertionError):
        f(3)


Ge0Le2: TypeAlias = Annotated[int, Interval(ge=0, le=2)]


def test_Interval_ge_le():
    @annotated
    def f(_: Ge0Le2):
        pass

    f(0)
    f(2)

    with pytest.raises(AssertionError):
        f(-1)
    with pytest.raises(AssertionError):
        f(3)


Mult3: TypeAlias = Annotated[int, MultipleOf(3)]


def test_MultipleOf():
    @annotated
    def f(_: Mult3):
        pass

    f(-3)
    f(0)
    f(3)
    f(6)

    with pytest.raises(AssertionError):
        f(-1)
    with pytest.raises(AssertionError):
        f(1)


NotEmpty: TypeAlias = Annotated[str, MinLen(1)]


def test_MinLen():
    @annotated
    def f(_: NotEmpty):
        pass

    f("a")
    with pytest.raises(AssertionError):
        f("")


SmallStr: TypeAlias = Annotated[str, MaxLen(3)]


def test_MaxLen():
    @annotated
    def f(_: SmallStr):
        pass

    f("")
    f("aaa")

    with pytest.raises(AssertionError):
        f("aaaa")


BigramOrTrigram: TypeAlias = Annotated[str, Len(2, 3)]


def test_Len():
    @annotated
    def f(_: BigramOrTrigram):
        pass

    f("aa")
    f("aaa")

    with pytest.raises(AssertionError):
        f("")
    with pytest.raises(AssertionError):
        f("aaaa")


TzAware_Ellipsis: TypeAlias = Annotated[datetime, Timezone(...)]


def test_Timezone_datetime_Ellipsis():
    @annotated
    def f(_: TzAware_Ellipsis):
        pass

    f(datetime.now(tz=ZoneInfo("Europe/Paris")))

    with pytest.raises(AssertionError):
        f(datetime.now(tz=None))


NaiveDatetime: TypeAlias = Annotated[datetime, Timezone(None)]


def test_Timezone_datetime_None():
    @annotated
    def f(_: NaiveDatetime):
        pass

    f(datetime.now(tz=None))

    with pytest.raises(AssertionError):
        f(datetime.now(timezone.utc))


AtParisTz_str: TypeAlias = Annotated[datetime, Timezone("Europe/Paris")]


def test_Timezone_datetime_str():
    @annotated
    def f(_: AtParisTz_str):
        pass

    f(datetime.now(tz=ZoneInfo("Europe/Paris")))

    with pytest.raises(AssertionError):
        f(datetime.now(timezone.utc))


AtParisTz_ZoneInfo: TypeAlias = Annotated[datetime, Timezone(ZoneInfo("Europe/Paris"))]


def test_Timezone_datetime_ZoneInfo():
    @annotated
    def f(_: AtParisTz_ZoneInfo):
        pass

    f(datetime.now(tz=ZoneInfo("Europe/Paris")))

    with pytest.raises(AssertionError):
        f(datetime.now(timezone.utc))


def test_predicate_input():
    @annotated
    def f(_: LowerCase):
        pass

    f("aaaa")
    with pytest.raises(AssertionError):
        f("aaaA")


def test_predicate_return():
    @annotated
    def f(a: str) -> LowerCase:
        return a

    f("aaaa")
    with pytest.raises(AssertionError):
        f("aaaA")


Documented: TypeAlias = Annotated[int, doc("Well, it's a integer")]


def test_Doc():
    @annotated
    def f(a: Documented) -> Documented:
        return a

    assert f(1) == 1
    with pytest.raises(AssertionError):
        f("aaaA")  # type: ignore


def test_list_of_annotated():
    @annotated
    def f(a: Annotated[list[Documented], MinLen(1)]) -> Documented:
        return a[0]

    assert f([1, 2]) == 1
    with pytest.raises(AssertionError):
        f(None)  # type: ignore
    with pytest.raises(AssertionError):
        f("aaaA")  # type: ignore
    with pytest.raises(AssertionError):
        f([1, "aaaA"])  # type: ignore


def test_tuple_of_length_ge_1():
    @annotated
    def f(a: Annotated[tuple[Documented, ...], MinLen(1)]) -> Documented:
        return a[0]

    assert f((1,)) == 1
    assert f((1, 2)) == 1
    with pytest.raises(AssertionError):
        f(tuple())  # type: ignore
    with pytest.raises(AssertionError):
        f([1, 2])  # type: ignore


def test_empty_tuple():
    @annotated
    def f(a: tuple[()]):
        pass

    assert f(tuple()) is None
    with pytest.raises(AssertionError):
        f(None)  # type: ignore
    with pytest.raises(AssertionError):
        f((1,))  # type: ignore


def test_list_of_tuple_of_annotated():
    @annotated
    def f(a: Annotated[list[tuple[Documented, Documented]], MinLen(1)]) -> Documented:
        return a[0][0]

    assert f([(1, 2)]) == 1

    with pytest.raises(AssertionError):
        f([(1,)])  # type: ignore
    with pytest.raises(AssertionError):
        f([(1, 2.0)])  # type: ignore


def test_list_of_tuple_of_constrained():
    @annotated
    def f(a: Annotated[list[tuple[Annotated[int, Gt(0)], int]], MinLen(1)]) -> Documented:
        return a[0][0]

    assert f([(1, 0)]) == 1

    with pytest.raises(AssertionError):
        f([(0, 0)])  # type: ignore


def test_tuple_ellipsis():
    @annotated
    def f(a: tuple[int, ...]):
        return a[-1]

    assert f((1, 0)) == 0

    with pytest.raises(AssertionError):
        f((0, 1.0))  # type: ignore


type IntGt0 = Annotated[int, Gt(0)]


def test_type_alias():
    @annotated
    def f(a: IntGt0):
        return a

    assert f(1) == 1
    with pytest.raises(AssertionError):
        f(1.0)  # type: ignore
    with pytest.raises(AssertionError):
        f(0)  # type: ignore


def test_union_types():
    @annotated
    def f(a: int | None):
        return a

    assert f(None) is None
    assert f(0) == 0
    with pytest.raises(AssertionError):
        f(1.0)  # type: ignore


type RecList = list[RecList]


def test_recursive_type():
    @annotated
    def f(a: RecList):
        actual = a
        while actual:
            actual = actual[0]
        return actual

    e = []

    assert f([[[e]]]) is e
    assert f([[[e], e], e]) is e
    with pytest.raises(AssertionError):
        _ = f([[[e]], 1])  # type: ignore
    with pytest.raises(AssertionError):
        _ = f([[[e, None]]])  # type: ignore


type MutRec1 = list[MutRec2]
type MutRec2 = MutRec1 | int


def test_mutually_recursive_types():
    @annotated
    def f(a: MutRec1):
        actual = a
        while isinstance(actual, list):
            actual = actual[0]
        return actual

    assert f([[[0]]]) == 0


type PropsMap = dict[str, Annotated[float, Gt(0.0)]]


def test_dict():
    @annotated
    def get_val(kv: PropsMap, key: str, default: float = 0.0) -> float:
        if key in kv:
            return kv[key]
        else:
            return default

    assert get_val({"toto": 10.0}, "toto") == 10.0
    assert get_val({"toto": 10.0}, "tata") == 0.0

    with pytest.raises(AssertionError):
        _ = get_val({"toto": 10}, "toto")

    with pytest.raises(AssertionError):
        _ = get_val({"toto": 0.0}, "toto")


type Height = Annotated[float, Ge(0.0)]
type InternalPressureFlyingAtm = Annotated[float, Interval(ge=0.0, le=1.0)]
type Depth = Annotated[float, Le(0.0)]
type InternalPressureDivingAtm = Annotated[float, Ge(1.0)]
type FlyOrDivePressureAtm = tuple[Height, InternalPressureFlyingAtm] | tuple[
    Depth, InternalPressureDivingAtm
]


def test_advanced_union():
    @annotated
    def pressure_is_ok(p: FlyOrDivePressureAtm):  # pragma: no cover
        if p[0] == 0.0:
            # on the ground: pressure should be constant
            return p[1] == 1.0
        elif p[0] > 0.0:
            # flying: keep pressure high
            return p[1] >= 0.75
        elif p[0] < 0.0:
            # diving: keep pressure low
            return p[1] <= 1.25

    # at ground
    assert pressure_is_ok((0.0, 1.0))
    assert not pressure_is_ok((0.0, 0.5))
    assert not pressure_is_ok((0.0, 2.0))
    assert pressure_is_ok((-1.0, 1.0))
    assert pressure_is_ok((1.0, 1.0))

    with pytest.raises(AssertionError):  # type: ignore
        _ = pressure_is_ok((0.0, -1.0))

    # in flight
    assert pressure_is_ok((1000.0, 0.9))
    assert not pressure_is_ok((1000.0, 0.5))
    with pytest.raises(AssertionError):  # type: ignore
        _ = pressure_is_ok((1000.0, 1.1))

    # diving
    assert pressure_is_ok((-1000.0, 1.1))
    assert not pressure_is_ok((-1000.0, 1.5))
    with pytest.raises(AssertionError):  # type: ignore
        _ = pressure_is_ok((-1000.0, 0.9))



from typing import Callable
from unittest import expectedFailure

import pytest

from contractme.contracting import par_inner, show_source

"""
This tests mainly the heuristics for displaying a nice source, as
the source parsing sacrifices precision in favor of accepting many
difference callable type.
"""


@pytest.mark.parametrize(
    "i,expected",
    [
        ("1 + 2", "1 + 2"),
        ("1 + 2)", "1 + 2"),
        ("(1 + 2)", "1 + 2"),
        ("( (1 + 2)", "1 + 2"),
        ("( ( 1 + 2 ) ) ", "1 + 2"),
        ("1 + (2)", "1 + (2)"),
        ("1 + (2))", "1 + (2)"),
    ],
)
def test_par_inner(i: str, expected: str):
    assert par_inner(i) == expected


@pytest.mark.parametrize("f,expected", [((lambda a: a), "a")])
def test_show_source(f: Callable, expected: str):
    assert show_source(f) == expected
    assert show_source((lambda a: a)) == "a"

    # trying to add some confusion
    def lambdaa(f):
        return f

    assert show_source(lambdaa(lambda a: a)) == "a"
    assert show_source(lambda f: (lambda a: a)) == "lambda a: a"



import sys
import datetime
import zoneinfo
import functools
import operator
from enum import Enum, auto
from dataclasses import dataclass
import types
import typing
from typing import (
    Annotated,
    Any,
    Callable,
    Iterable,
    TypeAliasType,
    TypeVar,
    get_origin,
    get_args,
)
import annotated_types

assert sys.version_info >= (3, 12)

# Stolen from typeshed
type ClassInfo = type | types.UnionType | tuple[ClassInfo, ...]
type AnnotationForm = Any

type Constraint = functools.partial | Any
# More specifically: some are taken into account (annotated-types),
# some are not (open world: it's yours and IDC about it)
# maybe they could be redispatched to extensions in the future? closing the
# world so that any malformed constraint could be detected?
type NormalizedConstraint = Any
# partial functions are re-contracted to annotated-types constraints


def normalize_constraint(constraint: Constraint) -> NormalizedConstraint:
    if isinstance(constraint, functools.partial):
        # https://github.com/annotated-types/annotated-types?tab=readme-ov-file#gt-ge-lt-le
        match constraint.func, constraint.args:
            case operator.lt, (n,):
                n = annotated_types.Lt(n)
            case operator.le, (n,):
                n = annotated_types.Le(n)
            case operator.gt, (n,):
                n = annotated_types.Gt(n)
            case operator.ge, (n,):
                n = annotated_types.Ge(n)
            case _:  # pragma: no cover
                raise NotImplementedError(constraint)
        return n
    else:
        return constraint


def get_all_constraint_errors(val, constraints: list[Constraint]) -> list[NormalizedConstraint]:
    if not constraints:
        return []
    else:
        failing = []
        for constraint_denorm in constraints:
            constraint = normalize_constraint(constraint_denorm)
            match constraint:
                case annotated_types.Interval(gt=None, lt=None, ge=None, le=None):
                    # null range
                    check = True
                case annotated_types.Gt(gt=gt) | annotated_types.Interval(
                    gt=gt, lt=None, ge=None, le=None
                ):
                    check = val > gt
                case annotated_types.Lt(lt=lt) | annotated_types.Interval(
                    lt=lt, gt=None, ge=None, le=None
                ):
                    check = val < lt
                case annotated_types.Ge(ge=ge) | annotated_types.Interval(
                    ge=ge, lt=None, gt=None, le=None
                ):
                    check = val >= ge
                case annotated_types.Le(le=le) | annotated_types.Interval(
                    le=le, lt=None, ge=None, gt=None
                ):
                    check = val <= le
                case annotated_types.Interval(gt=gt, lt=lt, ge=None, le=None):
                    check = gt < val < lt
                case annotated_types.Interval(ge=ge, lt=lt, gt=None, le=None):
                    check = ge <= val < lt
                case annotated_types.Interval(gt=gt, le=le, ge=None, lt=None):
                    check = gt < val <= le
                case annotated_types.Interval(ge=ge, le=le, gt=None, lt=None):
                    check = ge <= val <= le
                case annotated_types.MultipleOf(multiple_of=multiple_of):
                    check = val % multiple_of == 0
                case annotated_types.MinLen(min_length=min_length) | annotated_types.Len(
                    min_length=min_length, max_length=None
                ):
                    check = min_length <= len(val)
                case annotated_types.MaxLen(max_length=max_length) | annotated_types.Len(
                    max_length=max_length, min_length=None
                ):
                    check = len(val) <= max_length
                case annotated_types.Len(min_length=min_length, max_length=max_length):
                    assert min_length is not None and max_length is not None
                    check = min_length <= len(val) <= max_length
                case annotated_types.Timezone(tz=tz):
                    if tz is None:
                        check = val.tzinfo is None
                    elif tz is Ellipsis:
                        check = val.tzinfo is not None
                    elif isinstance(tz, datetime.tzinfo):
                        check = val.tzinfo == tz
                    elif isinstance(tz, str):
                        check = val.tzinfo == zoneinfo.ZoneInfo(tz)
                    else:  # pragma: no cover
                        raise NotImplementedError(f"tz being {tz.__class__.__name__}")
                case annotated_types.Predicate(func=func):
                    check = func(val)
                case _:
                    check = True

            if not check:
                failing.append(constraint)
        return failing


def resolve_type(t: AnnotationForm) -> AnnotationForm:
    """
    Best-effort type resolution, this will remove as many
    type var or alias as possible.
    """
    alias = []
    while isinstance(t, (TypeVar, TypeAliasType)):
        if t in alias:
            # circular definition
            break
        else:
            alias.append(t)

        if isinstance(t, TypeAliasType):
            t = t.__value__
        elif isinstance(t, TypeVar):
            t = t.__bound__ or Any
    return t


class ChildMod(Enum):
    NONE = auto()
    REPEAT = auto()  # 0..n
    REPEAT_KV = auto()  # {0: a, 1: b ...}
    OPTION = auto()  # A | B | C


type ConstraintSet = list[Any]


@dataclass
class CompositeTypeInfo:
    has_origin: bool
    origin: ClassInfo
    constraints: ConstraintSet
    child_mod: ChildMod
    children: tuple[AnnotationForm, ...]

    def __init__(self, t: AnnotationForm):
        self.has_origin = True
        origin = get_origin(t) or t
        self.constraints = list()
        while origin is Annotated:
            self.constraints += get_args(t)[1:]
            t = resolve_type(get_args(t)[0])
            origin = get_origin(t) or t
        self.origin = origin
        args = get_args(t)
        if self.origin == tuple and len(args) == 2 and args[1] == Ellipsis:
            self.child_mod = ChildMod.REPEAT
            self.children = (args[0],)
        elif self.origin == list:
            assert (
                len(args) == 1
            ), "Cannot have list[T1, T2...] must either be list[T1] or list[tuple[T1, T2]]"
            self.child_mod = ChildMod.REPEAT
            self.children = args
        elif self.origin in (dict,):
            self.child_mod = ChildMod.REPEAT_KV
            self.children = args
        elif self.origin in (types.UnionType, typing.Union):
            self.child_mod = ChildMod.OPTION
            self.children = args
        else:
            self.child_mod = ChildMod.NONE
            self.children = args
            self.has_origin = isinstance(self.origin, type)


def get_recursive_composite_type_info(an: AnnotationForm) -> Iterable[CompositeTypeInfo]:
    to_visit = [an]

    while to_visit:
        t = resolve_type(to_visit.pop(0))
        r = CompositeTypeInfo(t)
        if r.child_mod == ChildMod.NONE:
            to_visit += r.children
        yield r


@dataclass
class ErrorsCollector:
    _context: list
    _errors: list[str]
    _local_error: bool

    def __init__(self):
        self._context = []
        self._errors = []
        self._local_error = False

    def set_context(self, type_info):
        self._context.append(type_info.origin)

    def _append_error(self, msg):
        self._errors.append(f"in {self._context}: {msg}")

    def check_cond(self, cond: bool, message: str):
        self._local_error = not cond
        if self._local_error:
            self._append_error(message)

    def is_ok(self) -> bool:
        return not self._errors

    def locally_ok(self) -> bool:
        return not self._local_error

    def maybe_first_error(self) -> str | None:
        return None if not self._errors else self._errors[0]

    def all_errors(self) -> list[str]:
        return self._errors

    def clear_errors(self) -> None:
        self._errors = []


def get_first_structure_error(value, an, recursive_type_info=None) -> str | None:
    struct = ErrorsCollector()

    if recursive_type_info is None:
        recursive_type_info = get_recursive_composite_type_info(an)

    check_values = [value]
    for type_info in recursive_type_info:
        # Struct check
        val = check_values.pop(0)
        struct.set_context(type_info)
        is_sequence = type_info.origin in (list, tuple)
        if type_info.child_mod == ChildMod.OPTION:
            found = False
            for c in type_info.children:
                err = get_first_structure_error(val, c)
                if not err:
                    found = True
                    break
            struct.check_cond(found, f"{val} does not match any of {type_info.children}")
        elif type_info.has_origin:
            struct.check_cond(
                isinstance(val, type_info.origin),
                f"expect {type_info.origin} got {val.__class__.__name__}",
            )
            if struct.is_ok():
                if type_info.child_mod == ChildMod.REPEAT:
                    assert len(type_info.children) == 1
                    for v in val:
                        err = get_first_structure_error(v, type_info.children[0])
                        struct.check_cond(not err, str(err))
                        if not struct.is_ok():
                            break
                elif type_info.child_mod == ChildMod.REPEAT_KV:
                    assert len(type_info.children) == 2
                    for v in val.items():
                        err = get_first_structure_error(v[0], type_info.children[0])
                        struct.check_cond(not err, str(err))
                        err = get_first_structure_error(v[1], type_info.children[1])
                        struct.check_cond(not err, str(err))
                        if not struct.is_ok():
                            break
                elif is_sequence:
                    # depth-first
                    struct.check_cond(
                        len(val) == len(type_info.children),
                        f"not the right number of children: "
                        f"expect {len(type_info.children)} got {len(val)}",
                    )
                    if struct.is_ok():
                        check_values = list(val) + check_values

        if not struct.is_ok():
            break
    assert not check_values, f"bug! values remains to be checked ({check_values}) for {value} {an}"
    return struct.maybe_first_error()


def get_constraints_errors(
    value, an, recursive_type_info: Iterable[CompositeTypeInfo] | None = None
):
    constraints = ErrorsCollector()

    def only_valid(rti: Iterable[CompositeTypeInfo]):
        # unimplemented
        yield from rti

    if recursive_type_info is None:
        recursive_type_info = get_recursive_composite_type_info(an)

    check_values = [value]
    for type_info in only_valid(recursive_type_info):
        # Constraints check
        val = check_values.pop(0)
        if type_info.child_mod == ChildMod.OPTION:
            constraint_failures = None
            for child in type_info.children:
                # check structure, again, very innefficient
                if get_first_structure_error(val, child):
                    continue
                err = get_constraints_errors(val, child)
                constraints.check_cond(not err, "tried option {child_type_info}: {err}")
                if not err:
                    # found a config that works
                    constraints.clear_errors()
                    break
        else:
            is_sequence = type_info.origin in (list, tuple)
            constraint_failures = get_all_constraint_errors(val, type_info.constraints)
            constraints.check_cond(
                not constraint_failures,
                f"{val!r} failed constraint {constraint_failures}",
            )
            if constraints.locally_ok():
                # check only if parent constraints are already ok
                if type_info.child_mod == ChildMod.REPEAT:
                    assert len(type_info.children) == 1
                    for v in val:
                        err = get_constraints_errors(v, type_info.children[0])
                        constraints.check_cond(not err, str(err))
                elif type_info.child_mod == ChildMod.REPEAT_KV:
                    assert len(type_info.children) == 2
                    for v in val.items():
                        err = get_constraints_errors(v[0], type_info.children[0])
                        constraints.check_cond(not err, str(err))
                        err = get_constraints_errors(v[1], type_info.children[1])
                        constraints.check_cond(not err, str(err))
                elif is_sequence:
                    # depth-first
                    check_values = list(val) + check_values

    assert not check_values, f"bug! values remains to be checked ({check_values}) for {value} {an}"
    return constraints.all_errors()
